/*
 * Copyright (C) 2022-02-19  Julien BRUGUIER
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <string>
#include <sstream>


#include <src/plugin.h>

extern "C"
{

void plugin_configure(void *plugin)
{
	::svm_plugin_configure(plugin,
		"PLUGIN coroutine \n"
		"DEFINE \n"
		"	OPTION coroutine.prefix -p STR \n"
		"	INSTRUCTION coroutine.call STR:name [ STR SYM ]:function PTR:parameters \n"
		"	INSTRUCTION coroutine.run STR:name \n"
		"	INSTRUCTION coroutine.yield \n"
		"	INSTRUCTION coroutine.return STR:name \n"
		,SVM_API_SIGNATURE,SVM_VERSION);
}

#ifndef ARGV_VALUE
#define ARGV_VALUE(index,type) ::svm_value_##type##_get(svm,::svm_parameter_value_get(svm,argv[(index)]))
#endif
#ifndef ARGV_PLUGIN
#define ARGV_PLUGIN(index,plugin,name) reinterpret_cast<type_##name*>(::svm_value_plugin_get_internal(svm,::svm_parameter_value_get(svm,argv[(index)])))
#endif
#ifndef ARGV_MARKER
#define ARGV_MARKER(index) std::string(::svm_parameter_marker_get(svm,argv[(index)]).string)
#endif
#ifndef ARGV_KEYWORD
#define ARGV_KEYWORD(index) std::string(::svm_parameter_keyword_get(svm,argv[(index)]).string)
#endif
#ifndef ARGV_STRUCT
#define ARGV_STRUCT(index,plugin,name) reinterpret_cast<struct_##name*>(::svm_structure_get_internal(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name),::svm_parameter_structure_get(svm,argv[(index)])))
#endif
#ifndef ARGV_VARIABLE
#define ARGV_VARIABLE(index) ::svm_parameter_variable_get(svm,argv[(index)]);
#endif
#ifndef NEW_VALUE
#define NEW_VALUE(type,value) ::svm_value_##type##_new(svm,(value))
#endif
#ifndef NEW_PLUGIN
#define NEW_PLUGIN(plugin,name,value) ::svm_value_plugin_new(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name),(value))
#endif
#ifndef NEW_STRUCT
#define NEW_STRUCT(plugin,name,value) ::svm_structure_new(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name),(value))
#endif
#ifndef NEW_STRING
#define NEW_STRING(raw_string) ::svm_string_new(svm,raw_string.c_str(),raw_string.size())
#endif
#ifndef RAW_STRING
#define RAW_STRING(svm_string) std::string(svm_string.string,svm_string.size)
#endif
#ifndef NEW_BOOLEAN
#define NEW_BOOLEAN(boolean) ((boolean)?TRUE:FALSE)
#endif
#ifndef RAW_BOOLEAN
#define RAW_BOOLEAN(boolean) ((boolean)==TRUE)
#endif
#ifndef NEW_NULL_VALUE
#define NEW_NULL_VALUE(type) ::svm_value_##type##_new_null(svm)
#endif
#ifndef NEW_NULL_PLUGIN
#define NEW_NULL_PLUGIN(plugin,name) ::svm_value_plugin_new_null(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name))
#endif
#ifndef NEW_NULL_STRUCT
#define NEW_NULL_STRUCT(plugin,name) ::svm_structure_new_null(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name))
#endif
#ifndef ERROR_INTERNAL
#define ERROR_INTERNAL(irq,message) ::svm_processor_current_raise_error_internal__raw(svm,irq,message)
#endif
#ifndef ERROR_EXTERNAL
#define ERROR_EXTERNAL(plugin,name,message) ::svm_processor_current_raise_error_external__raw(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name),message)
#endif
#ifndef CONST_PEP
#define CONST_PEP(plugin,name) ::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name)
#endif
#ifndef CURRENT
#define CURRENT(object) ::svm_##object##_get_current(svm)
#endif
#ifndef RETURN
#define RETURN return nullptr
#endif
#ifndef VARIABLE_GLOBAL
#define VARIABLE_GLOBAL(variable) ::svm_variable_scope_set_global(svm,(variable))
#endif
#ifndef VARIABLE_LOCAL
#define VARIABLE_LOCAL(variable) ::svm_variable_scope_set_local(svm,(variable))
#endif
#ifndef VARIABLE_DELETE
#define VARIABLE_DELETE(variable) ::svm_variable_delete(svm,(variable))
#endif

}

#line 139 "../plugins/coroutine/core/coroutine.svm_plugin"

namespace Coroutine
{

static std::string _prefix = "coroutine:";

std::string flag(const std::string& name)
{
	return _prefix+name;
}

std::pair<size_t,size_t> stack(const void *svm, SVM_Kernel kernel, const std::string& name)
{
	std::string flag = Coroutine::flag(name);
	auto level_begin = ::svm_processor_returnstack_find_flag__raw(svm,kernel,flag.c_str(),0);
	if(level_begin==::svm_processor_returnstack_get_size(svm,kernel))
	{
		std::ostringstream oss;
		oss << "Coroutine " << name << " does not exist.";
		ERROR_INTERNAL(FAILURE,oss.str().c_str());
	}
	auto level_end = level_begin;
	auto level = level_begin;
	do
	{
		level_end = level+1;
		level = ::svm_processor_returnstack_find_flag__raw(svm,kernel,flag.c_str(),level_end);
		
	} while(level!=::svm_processor_returnstack_get_size(svm,kernel));
	return std::make_pair(level_begin,level_end);
}
}
#line 150 "src/plugin.cpp"

extern "C"
{

/* OPTION coroutine.prefix -p STR */


/* INSTRUCTION coroutine.call STR:name [ STR SYM ]:function PTR:parameters */

SVM_Value instruction_call(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 190 "../plugins/coroutine/core/coroutine.svm_plugin"
	SVM_Kernel kernel = ::svm_kernel_get_current(svm);
	SVM_String raw_name = ARGV_VALUE(0,string);
	std::string name(raw_name.string,raw_name.size);
	std::string flag = Coroutine::flag(name);
	auto level = ::svm_processor_returnstack_find_flag__raw(svm,kernel,flag.c_str(),0);
	if(level!=::svm_processor_returnstack_get_size(svm,kernel))
	{
		std::ostringstream oss;
		oss << "Coroutine " << name << " already exists.";
		ERROR_INTERNAL(FAILURE,oss.str().c_str());
	}
	SVM_Value function = ::svm_parameter_value_get(svm,argv[1]);
	if(::svm_value_type_is_string(svm,function))
	{
		SVM_Address address = ::svm_code_label_get_address(svm,::svm_processor_get_currentcode(svm,kernel),function);
		::svm_processor_call_local(svm,kernel,address,::svm_parameter_value_get(svm,argv[2]));
	}
	else
	{
		::svm_processor_call_global(svm,kernel,function,::svm_parameter_value_get(svm,argv[2]));
	}
	::svm_processor_set_flag__raw(svm,kernel,flag.c_str(),CASCADE);
	::svm_processor_returnstack_swap_level(svm,kernel,0);
	::svm_processor_returnstack_move_level(svm,kernel,0,1,::svm_processor_returnstack_get_size(svm,kernel)-1);
#line 187 "src/plugin.cpp"
	return nullptr;
}


/* INSTRUCTION coroutine.run STR:name */

SVM_Value instruction_run(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 224 "../plugins/coroutine/core/coroutine.svm_plugin"
	SVM_Kernel kernel = ::svm_kernel_get_current(svm);
	SVM_String raw_name = ARGV_VALUE(0,string);
	std::string name(raw_name.string,raw_name.size);
	std::string flag = Coroutine::flag(name);
	auto coroutine = Coroutine::stack(svm,kernel,name);
	::svm_processor_returnstack_swap_level(svm,kernel,coroutine.first);
	::svm_processor_returnstack_move_level(svm,kernel,coroutine.first,coroutine.first+1,0);
	::svm_processor_returnstack_move_level(svm,kernel,coroutine.first+1,coroutine.second,0);
#line 205 "src/plugin.cpp"
	return nullptr;
}


/* INSTRUCTION coroutine.yield */

SVM_Value instruction_yield(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 242 "../plugins/coroutine/core/coroutine.svm_plugin"
	SVM_Kernel kernel = ::svm_kernel_get_current(svm);
	std::string flag;
	for(SVM_Value_String *it = ::svm_processor_list_flag(svm,kernel) ; *it ; ++it)
	{
		SVM_String raw_f = ::svm_value_string_get(svm,*it);
		std::string f(raw_f.string,raw_f.size);
		if(f.substr(0,Coroutine::_prefix.size())==Coroutine::_prefix)
		{
			flag = f;
			break;
		}
	}
	if(not flag.empty())
	{
		ssize_t level_end = -1;
		auto level = level_end;
		do
		{
			level_end = level+1;
			level = ::svm_processor_returnstack_find_flag__raw(svm,kernel,flag.c_str(),level_end);
		} while(level!=::svm_processor_returnstack_get_size(svm,kernel));
		::svm_processor_returnstack_move_level(svm,kernel,0,level_end,::svm_processor_returnstack_get_size(svm,kernel)-level_end);
		::svm_processor_returnstack_swap_level(svm,kernel,0);
		::svm_processor_returnstack_move_level(svm,kernel,0,1,::svm_processor_returnstack_get_size(svm,kernel)-level_end-1);
	}
#line 240 "src/plugin.cpp"
	return nullptr;
}


/* INSTRUCTION coroutine.return STR:name */

SVM_Value instruction_return(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 278 "../plugins/coroutine/core/coroutine.svm_plugin"
	SVM_Kernel kernel = ::svm_kernel_get_current(svm);
	SVM_String raw_name = ARGV_VALUE(0,string);
	std::string name(raw_name.string,raw_name.size);
	std::string flag = Coroutine::flag(name);
	for(SVM_Value_String *it = ::svm_processor_list_flag(svm,kernel) ; *it ; ++it)
	{
		SVM_String raw_f = ::svm_value_string_get(svm,*it);
		std::string f(raw_f.string,raw_f.size);
		if(f==flag)
		{
			ERROR_INTERNAL(FAILURE,"Cannot return a coroutine from itself.");
		}
	}
	auto coroutine = Coroutine::stack(svm,kernel,name);
	::svm_processor_returnstack_swap_level(svm,kernel,coroutine.first);
	::svm_processor_returnstack_move_level(svm,kernel,coroutine.first,coroutine.first+1,0);
	::svm_processor_returnstack_move_level(svm,kernel,coroutine.first+1,coroutine.second,0);
	for(size_t i=0 ; i<(coroutine.second-coroutine.first) ; ++i)
	{
		::svm_processor_return(svm,kernel);
	}
#line 271 "src/plugin.cpp"
	return nullptr;
}


/* Generic handling functions */

void plugin_initialisation(const void *svm)
{
#line 175 "../plugins/coroutine/core/coroutine.svm_plugin"
	SVM_Value_String rp = ::svm_plugin_get_option(svm,CONST_PEP(coroutine,prefix));
	if(not ::svm_value_state_is_null(svm,rp))
	{
		SVM_String sp = ::svm_value_string_get(svm,rp);
		Coroutine::_prefix = RAW_STRING(sp);
	}
#line 287 "src/plugin.cpp"
}

}
